@@ -20,6 +20,7 @@ http://www.verisign.com/nds/naming/namestore/techdocs.html
 package com.verisign.epp.serverstub;
 
 // Logging Imports
+import java.io.ByteArrayInputStream;
 import java.io.File;
 import java.io.FileInputStream;
 import java.io.InputStream;
@@ -35,10 +36,15 @@ import java.util.Enumeration;
 import java.util.GregorianCalendar;
 import java.util.List;
 import java.util.Random;
+import java.util.StringTokenizer;
 import java.util.Vector;
 
 import org.apache.commons.codec.binary.Base64;
 import org.apache.log4j.Logger;
+import org.apache.xerces.dom.DocumentImpl;
+import org.w3c.dom.Document;
+import org.xml.sax.SAXNotRecognizedException;
+import org.xml.sax.SAXNotSupportedException;
 
 import com.verisign.epp.codec.domain.EPPDomainCheckCmd;
 import com.verisign.epp.codec.domain.EPPDomainCreateCmd;
@@ -47,6 +53,8 @@ import com.verisign.epp.codec.domain.EPPDomainInfoCmd;
 import com.verisign.epp.codec.domain.EPPDomainInfoResp;
 import com.verisign.epp.codec.domain.EPPDomainPendActionMsg;
 import com.verisign.epp.codec.domain.EPPDomainUpdateCmd;
+import com.verisign.epp.codec.gen.EPPCodec;
+import com.verisign.epp.codec.gen.EPPEncodeException;
 import com.verisign.epp.codec.gen.EPPResponse;
 import com.verisign.epp.codec.gen.EPPResult;
 import com.verisign.epp.codec.gen.EPPTransId;
@@ -68,13 +76,16 @@ import com.verisign.epp.codec.mark.EPPMark;
 import com.verisign.epp.codec.mark.EPPMarkAddress;
 import com.verisign.epp.codec.mark.EPPMarkContact;
 import com.verisign.epp.codec.mark.EPPTrademark;
+import com.verisign.epp.codec.signedMark.EPPEncodedSignedMark;
 import com.verisign.epp.codec.signedMark.EPPSignedMark;
+import com.verisign.epp.codec.signedMark.SMDRevocationList;
 import com.verisign.epp.exception.EPPException;
 import com.verisign.epp.framework.EPPEvent;
 import com.verisign.epp.framework.EPPEventResponse;
 import com.verisign.epp.framework.EPPPollQueueException;
 import com.verisign.epp.framework.EPPPollQueueMgr;
 import com.verisign.epp.util.EPPCatFactory;
+import com.verisign.epp.util.EPPSchemaCachingParser;
 import com.verisign.epp.util.Environment;
 
 /**
@@ -93,6 +104,11 @@ public class LaunchDomainHandler extends DomainHandler {
 	 */
 	private PKIXParameters pkixParameters;
 
+	/**
+	 * SMD revocation list
+	 */
+	private SMDRevocationList smdRevocationList = new SMDRevocationList();
+
 	/**
 	 * Constructs an instance of LaunchDomainHandler
 	 */
@@ -110,10 +126,22 @@ public class LaunchDomainHandler extends DomainHandler {
 			System.exit(1);
 		}
 
-		String crl = Environment.getOption("EPP.SignedMark.crl");
+		cat.debug("LaunchDomainHandler.LaunchDomainHandler(): Signed Mark Truststore = " + truststore);
+
+		// Load the CRL's
+		String crlsProp = Environment.getOption("EPP.SignedMark.crls");
+
+		List<String> crls = new ArrayList<String>();
+
+		if (crlsProp != null) {
+			StringTokenizer tokenizer = new StringTokenizer(crlsProp, ",");
+			while (tokenizer.hasMoreTokens()) {
+				crls.add(tokenizer.nextToken());
+			}
+		}
 
 		try {
-			this.pkixParameters = this.loadPKIXParameters(truststore, crl);
+			this.pkixParameters = this.loadPKIXParameters(truststore, crls);
 		}
 		catch (Exception ex) {
 			cat.error("LaunchDomainHandler.LaunchDomainHandler(): Error loading the public key: "
@@ -122,6 +150,41 @@ public class LaunchDomainHandler extends DomainHandler {
 			System.exit(1);
 		}
 
+		// Initialize the SMD revocation list
+		String revocationListProp = Environment
+				.getOption("EPP.SignedMark.revocationList");
+
+		if (revocationListProp == null) {
+
+			File smdRevocationListFile = new File(revocationListProp);
+
+			if (smdRevocationListFile.exists()) {
+
+				try {
+					FileInputStream smdRevocationListStream = new FileInputStream(
+							smdRevocationListFile);
+					this.smdRevocationList.decode(smdRevocationListStream);
+				}
+				catch (Exception ex) {
+					cat.error("LaunchDomainHandler.LaunchDomainHandler(): Error loading SMD revocation list : "
+							+ ex);
+					ex.printStackTrace();
+					System.exit(1);
+				}
+
+				cat.debug("LaunchDomainHandler.LaunchDomainHandler(): No SMD revocation list found");
+
+			}
+			else {
+				cat.error("LaunchDomainHandler.LaunchDomainHandler(): Error finding SMD revocation list \""
+						+ revocationListProp + "\"");
+				System.exit(1);
+			}
+		}
+		else {
+			cat.debug("LaunchDomainHandler.LaunchDomainHandler(): EPP.SignedMark.revocationList property not defined");
+		}
+
 		cat.debug("LaunchDomainHandler.LaunchDomainHandler(): exit");
 	}
 
@@ -135,8 +198,8 @@ public class LaunchDomainHandler extends DomainHandler {
 	 * 
 	 * @param aTrustStoreName
 	 *            Trust store file name
-	 * @param aCrlName
-	 *            Certificate Revocation List (CRL) file name
+	 * @param aCrls
+	 *            List of Certificate Revocation List (CRL) file names
 	 * 
 	 * @return Initialized <code>PKIXParameters</code> instance.
 	 * 
@@ -144,48 +207,59 @@ public class LaunchDomainHandler extends DomainHandler {
 	 *             Error initializing the PKIX parameters
 	 */
 	private PKIXParameters loadPKIXParameters(String aTrustStoreName,
-			String aCrlName) throws Exception {
+			List<String> aCrls) throws Exception {
 		cat.debug("LaunchDomainHandler.loadPKIXParameters(String, String): enter");
 
 		KeyStore trustStore = KeyStore.getInstance(KeyStore.getDefaultType());
 		FileInputStream trustStoreFile = new FileInputStream(aTrustStoreName);
 		trustStore.load(trustStoreFile, null);
+		trustStoreFile.close();
+		cat.debug("LaunchDomainHandler.loadPKIXParameters(String, String): truststore = " + aTrustStoreName);
 		PKIXParameters pkixParameters = new PKIXParameters(trustStore);
 
-		// CRL not defined?
-		if (aCrlName == null) {
-			cat.debug("LaunchDomainHandler.loadPKIXParameters(String, String): No CRL specified");
-			pkixParameters.setRevocationEnabled(false);
-			cat.debug("LaunchDomainHandler.loadPKIXParameters(String, String): exit");
-			return pkixParameters;
-		}
-
-		File crlFile = new File(aCrlName);
-
-		if (!crlFile.exists()) {
-			throw new EPPException("CRL file " + crlFile + " does not exist.");
-		}
-
 		CertificateFactory certFactory = CertificateFactory
 				.getInstance("X.509");
+
 		Collection crlContentsList = new ArrayList();
-		InputStream inStream = null;
 
-		try {
-			inStream = new FileInputStream(aCrlName);
-			crlContentsList.add(certFactory.generateCRL(inStream));
-		}
-		finally {
-			if (inStream != null) {
-				inStream.close();
+		for (String currCrl : aCrls) {
+			File crlFile = new File(currCrl);
+			if (crlFile.exists()) {
+				InputStream inStream = null;
+
+				try {
+					cat.debug("LaunchDomainHandler.loadPKIXParameters(String, String): adding CRL " + currCrl);
+					inStream = new FileInputStream(currCrl);
+					crlContentsList.add(certFactory.generateCRL(inStream));
+				}
+				finally {
+					if (inStream != null) {
+						inStream.close();
+					}
+				}
+			}
+			else {
+				throw new EPPException("CRL file " + currCrl
+						+ " does not exist.");
 			}
-		}
-		List<CertStore> certStores = new ArrayList<CertStore>();
-		certStores.add(CertStore.getInstance("Collection",
-				new CollectionCertStoreParameters(crlContentsList)));
 
-		pkixParameters.setCertStores(certStores);
-		pkixParameters.setRevocationEnabled(true);
+		}
+
+		// At least 1 CRL was loaded
+		if (crlContentsList.size() != 0) {
+
+			List<CertStore> certStores = new ArrayList<CertStore>();
+			certStores.add(CertStore.getInstance("Collection",
+					new CollectionCertStoreParameters(crlContentsList)));
+
+			pkixParameters.setCertStores(certStores);
+			pkixParameters.setRevocationEnabled(true);
+			cat.debug("LaunchDomainHandler.loadPKIXParameters(String, String): Revocation enabled");
+		}
+		else {
+			pkixParameters.setRevocationEnabled(false);
+			cat.debug("LaunchDomainHandler.loadPKIXParameters(String, String): Revocation disabled");
+		}
 
 		cat.debug("LaunchDomainHandler.loadPKIXParameters(String, String): exit");
 		return pkixParameters;
@@ -249,19 +323,19 @@ public class LaunchDomainHandler extends DomainHandler {
 			if (!theLaunchCheck.hasType()
 					|| theLaunchCheck.getType().equals(
 							EPPLaunchCheck.TYPE_CLAIMS)) {
-				
+
 				cat.debug("LaunchDomainHandler.doDomainCheck: Launch Claims Check Form for Phase = "
 						+ theLaunchCheck.getPhase());
-				
+
 				// Get phase
 				String phase = theLaunchCheck.getPhase().getPhase();
 
 				// Claims Check Command?
 				if (phase.equals(EPPLaunchPhase.PHASE_CLAIMS)) {
 
-					EPPResponse theResponse = new EPPResponse(
-							new EPPTransId(theCommand.getTransId(), "54321-XYZ"));
-					
+					EPPResponse theResponse = new EPPResponse(new EPPTransId(
+							theCommand.getTransId(), "54321-XYZ"));
+
 					EPPLaunchChkData theExt = new EPPLaunchChkData(
 							theLaunchCheck.getPhase());
 
@@ -276,18 +350,15 @@ public class LaunchDomainHandler extends DomainHandler {
 
 						// Is there a matching mark for domain name?
 						if (exists) {
-							theExt
-									.addCheckResult(new EPPLaunchCheckResult(
-											domainName,
-											true,
-											Base64.encodeBase64String(domainName
+							theExt.addCheckResult(new EPPLaunchCheckResult(
+									domainName, true, Base64
+											.encodeBase64String(domainName
 													.getBytes())));
 							exists = !exists;
 						} // No matching mark for domain name
 						else {
-							theExt
-									.addCheckResult(new EPPLaunchCheckResult(
-											domainName, false));
+							theExt.addCheckResult(new EPPLaunchCheckResult(
+									domainName, false));
 						}
 					}
 
@@ -585,6 +656,8 @@ public class LaunchDomainHandler extends DomainHandler {
 
 				} // Signed marks passed?
 				else if (theCmdExt.hasSignedMarks()) {
+					
+					
 
 					List<EPPSignedMark> signedMarks = theCmdExt
 							.getSignedMarks();
@@ -597,27 +670,98 @@ public class LaunchDomainHandler extends DomainHandler {
 								theCommand.getTransId(),
 								"No signed marks provided");
 					}
+					
+					// If signed marks were passed via XML instead of Base64
+					// encoded?
+					if (!(signedMarks.get(0) instanceof EPPEncodedSignedMark)
+							&& (aData != null)
+							&& (aData instanceof SessionData)
+							&& ((SessionData) aData).getAttribute("PACKET") != null) {
+						// Re-parse the XML without normalization and set the
+						// signed marks for validation
+						EPPSchemaCachingParser theParser = new EPPSchemaCachingParser();
+						try {
+							theParser.setFeature(
+									EPPSchemaCachingParser.NORMALIZE_DATA,
+									false);
+							
+							SessionData sessionData = (SessionData) aData;
+														
+							byte[] packet = (byte []) sessionData.getAttribute("PACKET");
+							
+							cat.debug("Command with signed mark = [" + new String(packet) + "]");
+
+							Document theDoc = theParser
+									.parse(new ByteArrayInputStream(packet));
+							EPPDomainCreateCmd theCreateCmd = (EPPDomainCreateCmd) EPPCodec
+									.getInstance().decode(theDoc);
+
+							signedMarks = ((EPPLaunchCreate) theCreateCmd
+									.getExtension(EPPLaunchCreate.class))
+									.getSignedMarks();
+						}
+						catch (Exception e) {
+							cat.error("LaunchDomainHandler.doDomainCreate: Error decoding XML signed mark: "
+									+ e);
+							return this.returnError(EPPResult.COMMAND_FAILED,
+									theCommand.getTransId(),
+									"Error decoding XML signed mark");
+						}
+
+					}
 
 					// For each signed mark
 					for (EPPSignedMark signedMark : signedMarks) {
 
-						// Let's print out the contents in XML no matter what
-						// was passed in.
-						signedMark.setBase64Encode(false);
-
+						
 						cat.debug("LaunchDomainHandler.doDomainCreate: signedMark = ["
 								+ signedMark + "]");
+						
+						Document doc = new DocumentImpl();
+						String signedMarkXML;
+						try {
+							signedMarkXML = EPPUtil.toStringNoIndent(signedMark
+									.encode(doc));
+						}
+						catch (EPPEncodeException e) {
+							cat.error("LaunchDomainHandler.doDomainCreate: Error re-encoding signed mark: "
+									+ e);
+							return this.returnError(EPPResult.COMMAND_FAILED,
+									theCommand.getTransId(),
+									"Error re-encoding signed mark");
+						}
+						System.out
+								.println("\nLaunchDomainHandler.doDomainCreate: signed mark = \n"
+										+ signedMarkXML);
+						System.out
+								.println("\nLaunchDomainHandler.doDomainCreate: base64 signed mark = \n"
+										+ new String(Base64.encodeBase64(
+												signedMarkXML.getBytes(), true)));						
 
 						// Signature valid?
 						if (signedMark.validate(this.pkixParameters)) {
 							cat.debug("LaunchDomainHandler.doDomainCreate: Signature is valid");
 						}
 						else {
+							cat.error("LaunchDomainHandler.doDomainCreate: Signature is NOT valid");
 							return this.returnError(
-									EPPResult.PARAM_OUT_OF_RANGE,
+									EPPResult.PARAM_VALUE_POLICY_ERROR,
 									theCommand.getTransId(),
 									"Signed mark signature invalid");
 						}
+						
+						// SMD is not revoked?
+						if (!this.smdRevocationList.isRevoked(signedMark)) {
+							cat.debug("LaunchDomainHandler.doDomainCreate: signed mark is not revoked");
+						}
+						else {
+							cat.debug("LaunchDomainHandler.doDomainCreate: signed mark is revoked");
+							return this.returnError(
+									EPPResult.PARAM_VALUE_POLICY_ERROR,
+									theCommand.getTransId(),
+									"Signed mark is revoked");
+						}
+						
 
 					}
 
